@@ -10,6 +10,7 @@ from pathlib import Path
 from os import PathLike
 from random import choices
 from string import ascii_letters, digits
+from hashlib import sha1
 
 import numpy as np
 import pandas as pd
@@ -261,13 +262,18 @@ class MuData:
         )
 
     def _check_duplicated_attr_names(self, attr: str):
-        if any([not getattr(self.mod[mod_i], attr + "_names").is_unique for mod_i in self.mod]):
+        if any(
+            [
+                not getattr(self.mod[mod_i], attr + "_names").astype(str).is_unique
+                for mod_i in self.mod
+            ]
+        ):
             # If there are non-unique attr_names, we can only handle outer joins
             # under the condition the duplicated values are restricted to one modality
             dups = [
                 np.unique(
                     getattr(self.mod[mod_i], attr + "_names")[
-                        getattr(self.mod[mod_i], attr + "_names").duplicated()
+                        getattr(self.mod[mod_i], attr + "_names").astype(str).duplicated()
                     ]
                 )
                 for mod_i in self.mod
@@ -300,6 +306,35 @@ class MuData:
                 return True
         return False
 
+    def _check_changed_attr_names(self, attr: str):
+        attrhash = f"_{attr}hash"
+        attr_names_changed, attr_columns_changed = False, False
+        if not hasattr(self, attrhash):
+            attr_names_changed, attr_columns_changed = True, True
+        else:
+            for m, mod in self.mod.items():
+                if m in getattr(self, attrhash):
+                    cached_hash = getattr(self, attrhash)[m]
+                    new_hash = (
+                        sha1(
+                            np.ascontiguousarray(getattr(self.mod[m], attr).index.values)
+                        ).hexdigest(),
+                        sha1(
+                            np.ascontiguousarray(getattr(self.mod[m], attr).columns.values)
+                        ).hexdigest(),
+                    )
+                    if cached_hash[0] != new_hash[0]:
+                        attr_names_changed = True
+                        if not attr_columns_changed:
+                            attr_columns_changed = None
+                        break
+                    if cached_hash[1] != new_hash[1]:
+                        attr_columns_changed = True
+                else:
+                    attr_names_changed, attr_columns_changed = True, None
+                    break
+        return (attr_names_changed, attr_columns_changed)
+
     def copy(self, filename: Optional[PathLike] = None) -> "MuData":
         if not self.isbacked:
             mod = {}
@@ -358,6 +393,10 @@ class MuData:
         """Shape of data, all variables and observations combined (:attr:`n_obs`, :attr:`n_var`)."""
         return self.n_obs, self.n_vars
 
+    def __len__(self) -> int:
+        """Length defined as a total number of observations (:attr:`n_obs`)."""
+        return self.n_obs
+
     # # Currently rely on AnnData's interface for setting .obs / .var
     # # This code implements AnnData._set_dim_df for another namespace
     # def _set_dim_df(self, value: pd.DataFrame, attr: str):
@@ -367,6 +406,22 @@ class MuData:
     #     setattr(self, f"_{attr}", value)
     #     AnnData._set_dim_index(self, value_idx, attr)
 
+    def _create_global_attr_index(self, attr: str, axis: int):
+        if axis == (1 - self._axis):
+            # Shared indices
+            modindices = [getattr(self.mod[m], attr).index for m in self.mod]
+            if all([modindices[i].equals(modindices[i + 1]) for i in range(len(modindices) - 1)]):
+                attrindex = modindices[0].copy()
+            attrindex = reduce(
+                np.union1d, [getattr(self.mod[m], attr).index.values for m in self.mod]
+            )
+        else:
+            # Modality-specific indices
+            attrindex = np.concatenate(
+                [getattr(self.mod[m], attr).index.values for m in self.mod], axis=0
+            )
+        return attrindex
+
     def _update_attr(self, attr: str, axis: int, join_common: bool = False):
         """
         Update global observations/variables with observations/variables for each modality.
@@ -379,24 +434,10 @@ class MuData:
 
         prev_index = getattr(self, attr).index
 
-        # # No _mod_index when upon read
-        # # No _mod_index in mudata < 0.1.2
-        # attr_names_maybe_changed = False
-        # if not hasattr(self, "_mod_index"):
-        #     attr_names_maybe_changed = True
-        # else:
-        #     for m, mod in self.mod.items():
-        #         if m in self._mod_index:
-        #             if attr in self._mod_index[m]:
-        #                 if not getattr(self.mod[m], attr).index.equals(self._mod_index[m][attr]):
-        #                     attr_names_maybe_changed = True
-        #                     break
-        #             else:
-        #                 attr_names_maybe_changed = True
-        #                 break
-        #         else:
-        #             attr_names_maybe_changed = True
-        #             break
+        # # No _attrhash when upon read
+        # # No _attrhash in mudata < 0.2.0
+        _attrhash = f"_{attr}hash"
+        attr_changed = self._check_changed_attr_names(attr)
 
         attr_duplicated = self._check_duplicated_attr_names(attr)
         attr_intersecting = self._check_intersecting_attr_names(attr)
@@ -405,6 +446,10 @@ class MuData:
             warnings.warn(
                 f"{attr}_names are not unique. To make them unique, call `.{attr}_names_make_unique`."
             )
+            if self._axis == -1:
+                warnings.warn(
+                    f"Behaviour is not defined with axis=-1, {attr}_names need to be made unique first."
+                )
 
         # Check if the are same obs_names/var_names in different modalities
         # If there are, join_common=True request can not be satisfied
@@ -415,6 +460,10 @@ class MuData:
                 )
                 join_common = False
 
+        if not any(attr_changed):
+            # Nothing to update
+            return
+
         # Figure out which global columns exist
         columns_global = getattr(self, attr).columns[
             list(
@@ -455,52 +504,68 @@ class MuData:
             )
             data_global = data_global.loc[:, [c not in columns_common for c in data_global.columns]]
 
+        # TODO: take advantage when attr_changed[0] == False — only new columns to be added
+
         #
         # Join modality .obs/.var tables
         #
-        # Main case: no duplicates and no intersection
+        # Main case: no duplicates and no intersection if the axis is not shared
         #
-        if not attr_intersecting and not attr_duplicated:
-            if join_common:
-                # We checked above that attr_names are guaranteed to be unique and thus are safe to be used for joins
+        if not attr_duplicated:
+            # Shared axis
+            if axis == (1 - self._axis) or self._axis == -1:
+                # We assume attr_intersecting and can't join_common
                 data_mod = pd.concat(
                     [
                         getattr(a, attr)
-                        .drop(columns_common, axis=1)
                         .assign(**{rowcol: np.arange(getattr(a, attr).shape[0])})
                         .add_prefix(m + ":")
                         for m, a in self.mod.items()
                     ],
                     join="outer",
-                    axis=axis,
+                    axis=1,
                     sort=False,
                 )
-                data_common = pd.concat(
-                    [getattr(a, attr)[columns_common] for m, a in self.mod.items()],
-                    join="outer",
-                    axis=0,
-                    sort=False,
-                )
-                data_mod = data_mod.join(data_common, how="left", sort=False)
             else:
-                data_mod = pd.concat(
-                    [
-                        getattr(a, attr)
-                        .assign(**{rowcol: np.arange(getattr(a, attr).shape[0])})
-                        .add_prefix(m + ":")
-                        for m, a in self.mod.items()
-                    ],
-                    join="outer",
-                    axis=axis,
-                    sort=False,
-                )
+                if join_common:
+                    # We checked above that attr_names are guaranteed to be unique and thus are safe to be used for joins
+                    data_mod = pd.concat(
+                        [
+                            getattr(a, attr)
+                            .drop(columns_common, axis=1)
+                            .assign(**{rowcol: np.arange(getattr(a, attr).shape[0])})
+                            .add_prefix(m + ":")
+                            for m, a in self.mod.items()
+                        ],
+                        join="outer",
+                        axis=0,
+                        sort=False,
+                    )
+                    data_common = pd.concat(
+                        [getattr(a, attr)[columns_common] for m, a in self.mod.items()],
+                        join="outer",
+                        axis=0,
+                        sort=False,
+                    )
+                    data_mod = data_mod.join(data_common, how="left", sort=False)
 
-            # this occurs when join_common=True and we already have a global data frame, e.g. after reading from HDF5
-            if join_common:
-                sharedcols = data_mod.columns.intersection(data_global.columns)
-                data_global.rename(
-                    columns={col: f"global:{col}" for col in sharedcols}, inplace=True
-                )
+                    # this occurs when join_common=True and we already have a global data frame, e.g. after reading from H5MU
+                    sharedcols = data_mod.columns.intersection(data_global.columns)
+                    data_global.rename(
+                        columns={col: f"global:{col}" for col in sharedcols}, inplace=True
+                    )
+                else:
+                    data_mod = pd.concat(
+                        [
+                            getattr(a, attr)
+                            .assign(**{rowcol: np.arange(getattr(a, attr).shape[0])})
+                            .add_prefix(m + ":")
+                            for m, a in self.mod.items()
+                        ],
+                        join="outer",
+                        axis=0,
+                        sort=False,
+                    )
 
             for mod, amod in self.mod.items():
                 colname = mod + ":" + rowcol
@@ -512,7 +577,10 @@ class MuData:
                 col = col.astype(np.uint32)
                 data_mod.loc[:, colname] = col
 
-            if len(data_global) > 0:
+            if len(data_global.columns) > 0:
+                # TODO: if there were intersecting attrnames between modalities,
+                #       this will increase the size of the index
+                # Should we use attrmap to figure the index out?
                 data_mod = data_mod.join(data_global, how="left", sort=False)
 
         #
@@ -653,10 +721,6 @@ class MuData:
             attrm[mod] = mapping > 0
 
         now_index = getattr(self, attr).index
-        keep_index = prev_index.isin(now_index)
-        # keep_index = np.sum(list(mdict.values()), axis=0) > 0
-        # ^-- old map vs new map?!
-        new_index = ~now_index.isin(prev_index)
 
         if len(prev_index) == 0:
             # New object
@@ -664,38 +728,71 @@ class MuData:
         elif now_index.equals(prev_index):
             # Index is the same
             pass
-        elif len(now_index) != len(prev_index) and new_index.sum() == 0:
-            # Update .obsm/.varm (size might have changed)
-            for mx_key, mx in attrm.items():
-                if mx_key not in self.mod.keys():  # not a modality name
-                    attrm[mx_key] = attrm[mx_key][keep_index]
+        else:
+            keep_index = prev_index.isin(now_index)
+            new_index = ~now_index.isin(prev_index)
 
-            # Update .obsp/.varp (size might have changed)
-            for mx_key, mx in attrp.items():
-                attrp[mx_key] = attrp[mx_key][keep_index, keep_index]
-        elif len(now_index) == len(prev_index):
-            # NOTE: when new_index.sum() == 0,
-            # the old index wouldn't be reordered
-            if new_index.sum() == 0:
-                pass
-            else:  # renamed
-                # TODO: tru to use obsmap/varmap
+            if new_index.sum() == 0 or (
+                keep_index.sum() + new_index.sum() == len(now_index)
+                and len(now_index) > len(prev_index)
+            ):
+                # Another length (filtered) or new modality added
+                # Update .obsm/.varm (size might have changed)
+                # NOTE: .get_index doesn't work with duplicated indices
+                if any(prev_index.duplicated()):
+                    # Assume the relative order of duplicates hasn't changed
+                    # NOTE: .get_loc() for each element is too slow
+                    # We will rename duplicated in prev_index and now_index
+                    # in order to use .get_indexer
+                    # index_order = [
+                    #    prev_index.get_loc(i) if i in prev_index else -1 for i in now_index
+                    # ]
+                    prev_values = prev_index.values
+                    now_values = now_index.values
+                    for value in prev_index[np.where(prev_index.duplicated())[0]]:
+                        v_now = np.where(now_index == value)[0]
+                        v_prev = np.where(prev_index.get_loc(value))[0]
+                        for i in range(min(len(v_now), len(v_prev))):
+                            prev_values[v_prev[i]] = f"{str(value)}-{i}"
+                            now_values[v_now[i]] = f"{str(value)}-{i}"
+
+                    prev_index = pd.Index(prev_values)
+                    now_index = pd.Index(now_values)
+
+                index_order = prev_index.get_indexer(now_index)
+
+                for mx_key, mx in attrm.items():
+                    if mx_key not in self.mod.keys():  # not a modality name
+                        attrm[mx_key] = attrm[mx_key][index_order]
+                        attrm[mx_key][index_order == -1] = np.nan
+
+                # Update .obsp/.varp (size might have changed)
+                for mx_key, mx in attrp.items():
+                    attrp[mx_key] = attrp[mx_key][index_order, index_order]
+                    attrp[mx_key][index_order == -1, :] = -1
+                    attrp[mx_key][:, index_order == -1] = -1
+
+            elif len(now_index) == len(prev_index):
+                # Renamed since new_index.sum() != 0
                 # We have to assume the order hasn't changed
                 pass
-        else:
-            raise NotImplementedError(
-                f"{attr}_names seem to have been renamed and filtered at the same time. "
-                "There is no way to restore the order. MuData object has to be re-created from these modalities:\n"
-                "  mdata1 = MuData(mdata.mod)"
-            )
 
-        # # Write ._mod_index
-        # if attr_names_maybe_changed:
-        #     for m, mod in self.mod.items():
-        #         if not hasattr(self, "_mod_index"):
-        #             self._mod_index = dict()
-        #         self._mod_index[m] = self._mod_index.get(m, {})
-        #         self._mod_index[m][attr] = getattr(mod, attr).index
+            else:
+                raise NotImplementedError(
+                    f"{attr}_names seem to have been renamed and filtered at the same time. "
+                    "There is no way to restore the order. MuData object has to be re-created from these modalities:\n"
+                    "  mdata1 = MuData(mdata.mod)"
+                )
+
+        # Write _attrhash
+        if attr_changed:
+            if not hasattr(self, _attrhash):
+                setattr(self, _attrhash, dict())
+            for m, mod in self.mod.items():
+                getattr(self, _attrhash)[m] = (
+                    sha1(np.ascontiguousarray(getattr(mod, attr).index.values)).hexdigest(),
+                    sha1(np.ascontiguousarray(getattr(mod, attr).columns.values)).hexdigest(),
+                )
 
     def _shrink_attr(self, attr: str, inplace=True) -> pd.DataFrame:
         """
@@ -799,7 +896,8 @@ class MuData:
         """
         Update .obs slot of MuData with the newest .obs data from all the modalities
         """
-        self._update_attr("obs", axis=1, join_common=bool(True * self.axis == 1))
+        join_common = self.axis == 1
+        self._update_attr("obs", axis=1, join_common=join_common)
 
     def obs_names_make_unique(self):
         """
@@ -891,7 +989,8 @@ class MuData:
         """
         Update .var slot of MuData with the newest .var data from all the modalities
         """
-        self._update_attr("var", axis=0, join_common=bool(True * self.axis == 0))
+        join_common = self.axis == 0
+        self._update_attr("var", axis=0, join_common=join_common)
 
     def var_names_make_unique(self):
         """
@@ -1124,7 +1223,7 @@ class MuData:
                     )
                     if any(global_keys):
                         descr += f"\n  {attr}:\t{str([keys[i] for i in range(len(keys)) if global_keys[i]])[1:-1]}"
-        descr += f"\n  {len(self.mod)} modalities"
+        descr += f"\n  {len(self.mod)} modalit{'y' if len(self.mod) == 1 else 'ies'}"
         for k, v in self.mod.items():
             descr += f"\n    {k}:\t{v.n_obs} x {v.n_vars}"
             for attr in [
